/****************************************************************************
 *
 * Palm Universal Wireless Keyboard(TM) add-on driver for Xwindow. v1.0
 * Copyright GPL Licence. 
 * 2005.06.04 lampo1 <lampo1@paran.com>
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#define MAXNAME 16 
#define SHIFT 50    /* most linux systems assign keycode 50 to Shift_L */

/* One IR key can have multiple meaning by combinations */
/* with other modifier keys. */

struct keymapping {
    unsigned char irbyte;

    char nml[MAXNAME];
    unsigned int kcode_nml;

    char lfn[MAXNAME];
    unsigned int kcode_lfn;
};  

struct keymapping KeyTable[64] = {
    {0x0e,        "grave",0,           "",0},
    {0x16,            "1",0,           "",0},
    {0x1e,            "2",0,           "",0},
    {0x26,            "3",0,           "",0},
    {0x25,            "4",0,           "",0},
    {0x2e,            "5",0,           "",0},
    {0x36,            "6",0,           "",0},
    {0x3d,            "7",0,           "",0},
    {0x3e,            "8",0,           "",0},
    {0x46,            "9",0,           "",0},
    {0x45,            "0",0,           "",0},
    {0x4e,        "minus",0,           "",0},
    {0x55,        "equal",0,           "",0},
    {0x66,    "BackSpace",0,           "",0},
    {0x0d,          "Tab",0,           "",0},
    {0x15,            "q",0,           "",0},
    {0x1d,            "w",0,           "",0},
    {0x24,            "e",0,           "",0},
    {0x2d,            "r",0,           "",0},
    {0x2c,            "t",0,           "",0},
    {0x35,            "y",0,           "",0},
    {0x3c,            "u",0,           "",0},
    {0x43,            "i",0,           "",0},
    {0x44,            "o",0,           "",0},
    {0x4d,            "p",0,           "",0},
    {0x54,  "bracketleft",0,           "",0},
    {0x5b, "bracketright",0,           "",0}, 
    {0x5d,    "backslash",0,           "",0},
    {0x58,    "Caps_Lock",0,           "",0},
    {0x1c,            "a",0,           "",0},
    {0x1b,            "s",0,           "",0},
    {0x23,            "d",0,           "",0},
    {0x2b,            "f",0,           "",0},
    {0x34,            "g",0,           "",0},
    {0x33,            "h",0,           "",0},
    {0x3b,            "j",0,           "",0},
    {0x42,            "k",0,           "",0},
    {0x4b,            "l",0,           "",0},
    {0x4c,    "semicolon",0,           "",0},
    {0x52,   "apostrophe",0,           "",0},
    {0x5a,       "Return",0,           "",0},
    {0x12,      "Shift_L",0,           "",0},
    {0x1a,            "z",0,           "",0},
    {0x22,            "x",0,           "",0},
    {0x21,            "c",0,           "",0},
    {0x2a,            "v",0,           "",0},
    {0x32,            "b",0,           "",0},
    {0x31,            "n",0,           "",0},
    {0x3a,            "m",0,           "",0},
    {0x48,        "comma",0,           "",0},
    {0x49,       "period",0,           "",0},
    {0x28,           "Up",0,      "Prior",0},
    {0x59,      "Shift_R",0,           "",0},
    {0x4a,        "slash",0,           "",0},
    {0x14,    "Control_L",0,           "",0},
    {0x03,       "Escape",0,           "",0}, /* Cmd */
    {0x29,        "space",0,           "",0},
    {0x30,        "Alt_R",0,           "",0}, /* Alt_Gr */
    {0x5e,         "Left",0,       "Home",0},
    {0x60,         "Down",0,       "Next",0},
    {0x2f,        "Right",0,        "End",0},
    {0x1f,       "Delete",0,           "",0}, /* 62th */ 
    {0x02,             "",0,           "",0}, /* Fn_L */
    {0x11,             "",0,           "",0}  /* Cancel */
};

int fd;
unsigned int auto_off_time = 300;    /* seconds */
unsigned int typematic_rate = 1;     /* n times of unit cycle */
unsigned int typematic_delay = 3;    /* n times of unit cycle */
unsigned int lfn_FLAG = 0;
Display *d;

void alarm_inthandler(int fd) {
    close(fd);
    exit(EXIT_SUCCESS);
}

void read_keymap() {
    FILE *fp;
    int i;
    int len = 254;
    char buf[255];
    char cset[] = " \n";
    char *p0, *p1, *p2;
 
    fp = popen("xmodmap -pke", "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot execute xmodmap -pke\n");
        exit(EXIT_FAILURE);
    }

    while ( fgets(buf, len, fp) ) {

	/* parsing one line of the result of xmodmap -pke */
	strtok(buf, cset);          /* discard "keycode" */
	p0 = strtok(NULL, cset);    /* point to keycode */
	strtok(NULL, cset);         /* discard "="  */
	p1 = strtok(NULL, cset);    /* point to X keysym name */ 
	p2 = strtok(NULL, cset);    /* point to the shifted X keysym name */

	for ( i = 0 ; i <= 61; i++) {

	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].nml, p1)) ) {
		KeyTable[i].kcode_nml = atoi(p0);
	    }
	    
	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].lfn, p1)) ) {
		KeyTable[i].kcode_lfn = atoi(p0);
	    }
	}
    } 
    pclose(fp);

    #ifdef DEBUG
    printf("In printing...\n");
    for ( i = 0; i <= 63; i++ ) {
	printf("#%d %x %s %d %s %d\n",
	       i,
	       KeyTable[i].irbyte,
	       KeyTable[i].nml,
	       KeyTable[i].kcode_nml,
	       KeyTable[i].lfn,
	       KeyTable[i].kcode_lfn
	       );
    }
    #endif	       
}
    
void read_rcfile(FILE *fp) {

    char dmy[255];    
    while ( (fgets(dmy, sizeof(dmy), fp)) != NULL ) {

	if ( dmy[0] == '#' ) continue;
	else if ( dmy[0] == 'd' ) {
	    sscanf(dmy, "%*s %d", &typematic_delay);
	} else if ( dmy[0] == 'r' ) {
	    sscanf(dmy, "%*s %d", &typematic_rate);
	} else if ( dmy[0] == 'a' ) {
	    sscanf(dmy, "%*s %d", &auto_off_time);
	} 
    }
    fclose(fp);
}

void translate( unsigned char ch) {

    int  i;
    unsigned char msb;   /* 0 : key pressed */
                         /* 1 means key released. */

    msb =  (ch & 128) >> 7 ;   
    ch &= 127 ; 

    for (i = 0; i <= 64 ; i++) {

	if ( ch == KeyTable[i].irbyte ) {
	    if ( ch == 0x02 ) {         /* LeftFn */
		lfn_FLAG = (1^msb);
		break;
	    } else if ( lfn_FLAG == 1 ) {
		if ( ch == 0x1f ) {          /* LeftFn-Del */
		    close(fd);
		    exit(EXIT_SUCCESS);
		}
		if ( (ch == 0x55) || (ch == 0xd5) ) {       /* keymap reload */          
		    read_keymap();                               /* LeftFn-equal */
		    return;
		}

		XTestFakeKeyEvent(d, KeyTable[i].kcode_lfn, (1^msb), 0);
		XFlush(d);

                #ifdef DEBUG
		printf("code= %d\n", KeyTable[i].kcode_lfn);
                #endif


		break;		    
	    } else {
		XTestFakeKeyEvent(d, KeyTable[i].kcode_nml, (1^msb), 0);
		XFlush(d);

                #ifdef DEBUG
		printf("code= %d\n", KeyTable[i].kcode_nml);
                #endif

		break;
	    }
	}
    }
}
	
int main(int argc, char *argv[]) {

    unsigned char ch, irbyte;
    int nread, option;
    FILE *fp;
    char *device = "/dev/ttyS1";
    char *rcfile = "/etc/uwkrc";

    struct termios oldtio, newtio;
    struct sigaction newhandler;
    sigset_t blocked;

    while ( (option = getopt(argc, argv, "hd:c:")) != -1 ) {
        switch(option) {
        case 'h' :
            fprintf(stderr,"Palm Universal Wireless Keyboard driver v1.0\n");
            fprintf(stderr,"Usage : uwkxwin [-d device] [-c rcfile]\n");
            fprintf(stderr,"example : uwkxwin -d /dev/ttyS2 -c /etc/uwkrc\n");
            exit(0);
        case 'd' :
            device = optarg;
            break;
        case 'c' :
            rcfile = optarg;
            break;
        default :
            fprintf(stderr,"unknown option.\n");
            exit(1);
        }
    }
            

    /* signal install */
    newhandler.sa_handler = alarm_inthandler;
    sigemptyset(&blocked);
    sigaddset(&blocked, SIGALRM);
    newhandler.sa_mask = blocked;
    if ( sigaction(SIGALRM, &newhandler, NULL) == -1 ) {
	perror("sigaction install error\n");
	exit(EXIT_FAILURE);
    }

    /* open irda port device */
    fprintf(stderr, "Opening %s\n", device);
    fd = open(device, O_RDONLY | O_NOCTTY );
    if (fd == -1)  {
        fprintf(stderr, "Cannot open %s\n", device);
        exit(EXIT_FAILURE);
    }

    /* open rcfile */
    fprintf(stderr, "Opening %s\n", rcfile);
    fp = fopen( rcfile, "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot open %s\n", rcfile);
        fprintf(stderr,"Default time setting is applied.\n");
    } else {
	read_rcfile( fp );
    }

    /* read current xmodmap */
    read_keymap();

    /* Serial port setting to noncanonical, no echo, ...*/
    tcgetattr(fd, &oldtio); 
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
    newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 chars received */
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd,TCSANOW,&newtio);
               
    d = XOpenDisplay(NULL);
    
    while (1) {
	alarm( auto_off_time );

	nread = read(fd, &ch, 1);
        if ( ( nread > 0 ) && ( ch == 0xff ) ) {
      
	    nread = read(fd, &ch, 1);
	    if ( (nread > 0) && ( ch == 0xc0 ) ) {
	    
		nread = read(fd, &irbyte, 1);
		read(fd, &ch, 1);
		read(fd, &ch, 1); 
		nread = read(fd, &ch, 1);
 
		if ( ch == 0xc1 ) translate(irbyte);		
	    }
	} else if ( nread == -1 ) {
	    fprintf(stderr,"IrDA Port reading error.\n");
	    fprintf(stderr,"Recommend removing other irda modules.\n");
	    exit(EXIT_FAILURE);
	}         
    }  
}
